/// \file spiplayer.c Low-power SPI FLASH player
// derived from: spiusb.c Winbond 25X16 SPI flash/USB/Vorbis player
/*
  
  Copyright 2008-2010 VLSI Solution Oy. Absolutely no warranty.

*/  

//#define USE_WAV /* Play linear wav files using CodMiniWav from libdev1000. */
#define POFFTIME 60 /*5 minutes*/

/*Turn analog off in low-power pause mode to save some power. */
//#define PATCH_LOW_POWER_MODE

#include "system.h"

#define SPI_CLOCK_DIVIDER 2

// Command set of the SPI flash
#define SPI_EEPROM_COMMAND_WRITE_ENABLE  0x06
#define SPI_EEPROM_COMMAND_WRITE_DISABLE  0x04
#define SPI_EEPROM_COMMAND_READ_STATUS_REGISTER  0x05
#define SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER  0x01
#define SPI_EEPROM_COMMAND_READ  0x03
#define SPI_EEPROM_COMMAND_WRITE 0x02
#define SPI_EEPROM_COMMAND_CLEAR_ERROR_FLAGS 0x30
#define SPI_EEPROM_COMMAND_ERASE_4K_SECTOR 0x20
#define SPI_EEPROM_COMMAND_ERASE_CHIP 0xC7


// Position of workspace in Y ram, do not change.
#define WORKSPACE (mallocAreaY + 6144)


#include <stdio.h> //Standard io
#include <stdlib.h> // VS_DSP Standard Library
#include <vs1000.h> // VS1000B register definitions
#include <vectors.h> // VS1000B vectors (interrupts and services)
#include <minifat.h> // Read Only Fat Filesystem
#include <mapper.h> // Logical Disk
#include <string.h> // memcpy etc
#include <player.h> // VS1000B default ROM player
#include <audio.h> // DAC output
#include <codec.h> // CODEC interface
#include <vsNand.h>
#include <mappertiny.h>
#include <usb.h>

#include <dev1000.h>

#ifdef GPIO_CONTROL
#include "gpioctrl.h"
#endif /*GPIO_CONTROL*/

extern struct FsNandPhys fsNandPhys;
extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;
extern u_int16 codecVorbis[];

//macro to set SPI to MASTER; 8BIT; FSYNC Idle => xCS high
#define SPI_MASTER_8BIT_CSHI   PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1

//macro to set SPI to MASTER; 8BIT; FSYNC not Idle => xCS low
#define SPI_MASTER_8BIT_CSLO   PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE0

//macro to set SPI to MASTER; 16BIT; FSYNC not Idle => xCS low
#define SPI_MASTER_16BIT_CSLO  PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE0


/// Wait for not_busy (status[0] = 0) and return status
u_int16 SpiWaitStatus(void) {
  u_int16 status;
  SPI_MASTER_8BIT_CSHI;
  SPI_MASTER_8BIT_CSLO;
  SpiSendReceive(SPI_EEPROM_COMMAND_READ_STATUS_REGISTER);
  do {
    status = SpiSendReceive(0);
    if (PERIP(USB_STATUS) & USB_STF_BUS_RESET){
      USBHandler();
      SPI_MASTER_8BIT_CSHI;
      return -1; /* USB HAS BEEN RESET */
    }
  } while (status & 0x01);
    ; //Wait until chip is ready or return -1 if USB bus reset

  SPI_MASTER_8BIT_CSHI;
  
  return status;
}


// Block Read for SPI EEPROMS with 24-bit address e.g. up to 16MB 
u_int16 EeReadBlock(u_int16 blockn, u_int16 *dptr) {
  SpiWaitStatus();
  
  SPI_MASTER_8BIT_CSLO;
  SpiSendReceive(SPI_EEPROM_COMMAND_READ);
  SpiSendReceive(blockn>>7);            // Address[23:16] = blockn[14:7]
  SpiSendReceive((blockn<<1)&0xff);     // Address[15:8]  = blockn[6:0]0
  SpiSendReceive(0);                    // Address[7:0]   = 00000000
  SPI_MASTER_16BIT_CSLO;

  {
    int n;
    for (n=0; n<256; n++){
#if USE_INVERTED_DISK_DATA
      *dptr++ = ~SpiSendReceive(0);
#else
      *dptr++ = SpiSendReceive(0);
#endif
    }
  }
  SPI_MASTER_8BIT_CSHI;
  return 0;
}


struct FsMapper *FsMapSpiFlashCreate(struct FsPhysical *physical,
				u_int16 cacheSize);
s_int16 FsMapSpiFlashRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data);

const struct FsMapper spiFlashMapper = {
    0x010c, /*version*/
    256,    /*blocksize*/
    LOGICAL_DISK_BLOCKS,      /*blocks*/
    0,      /*cacheBlocks*/
    FsMapSpiFlashCreate,
    FsMapFlNullOk,//RamMapperDelete,
    FsMapSpiFlashRead,
    FsMapFlNullOk,//FsMapSpiFlashWrite,
    NULL,//FsMapFlNullOk,//RamMapperFree,
    FsMapFlNullOk,//FsMapSpiFlashFlush,//RamMapperFlush,
    NULL /* no physical */
};
struct FsMapper *FsMapSpiFlashCreate(struct FsPhysical *physical,
				     u_int16 cacheSize) {

  SPI_MASTER_8BIT_CSHI;
  PERIP(SPI0_FSYNC) = 0;
  PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (SPI_CLOCK_DIVIDER-1);
  PERIP(GPIO1_MODE) |= 0x1f; /* enable SPI pins */

  return &spiFlashMapper;
}



s_int16 FsMapSpiFlashRead(struct FsMapper *map, u_int32 firstBlock,
			  u_int16 blocks, u_int16 *data) {
  register s_int16 bl = 0;

  firstBlock += RESERVED_BLOCKS;

  while (bl < blocks) {
    EeReadBlock(firstBlock, data);
    data += 256;
    firstBlock++;
    bl++;
  }
  return bl;
}

#ifdef PATCH_LOW_POWER_MODE
void MyUSBSuspend(u_int16 timeOut) {
  PERIP(SCI_STATUS) |=   SCISTF_ANA_PDOWN | SCISTF_ANADRV_PDOWN;
  RealUSBSuspend(timeOut);
  PERIP(SCI_STATUS) &= ~(SCISTF_ANA_PDOWN | SCISTF_ANADRV_PDOWN);
}
#endif


void CheckSd(void) {
    if ((PERIP(GPIO0_DDR) & (1<<14)) == 0 &&
	(PERIP(GPIO0_IDATA) & (1<<14))) {
	/* SD not inserted */
    } else {
	/* SD inserted */
	//putch('S');
	putch('S');
	putch('D');
	putch('\n');

	Disable();
	ResetIRAMVectors();
	PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0;
	PERIP(INT_ENABLEH) = INTF_DAC;

	/* If SD inserted, run SD player. */
	PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
	SpiLoad(MMCSTART+4, 1/*24-bit address*/);
	while (1)
	    ;
    }
}

#ifdef USE_WAV
#include <codecminiwav.h>
#include <dev1000.h>
extern u_int16 codecVorbis[];
extern u_int16 ogg[];
extern u_int16 mInt[];

extern struct CodecServices cs;
extern struct Codec *cod;
enum CodecError PlayWavOrOggFile(void) {
  register enum CodecError ret = ceFormatNotFound;
  const char *eStr;
  LoadCheck(NULL, 0); /* higher clock, but 4.0x not absolutely required */
  if ((cod = CodMiniWavCreate())) {
      ret = cod->Decode(cod, &cs, &eStr);
      cod->Delete(cod);
#if 0
      if (earSpeakerReg)
      {
	  register int i;
	  for (i=0;i<DEFAULT_AUDIO_BUFFER_SAMPLES/(sizeof(tmpBuf)/2);i++) {
	      /* Must be memset each time, because effects like EarSpeaker are
		 performed in-place in the buffer */
	      memset(tmpBuf, 0, sizeof(tmpBuf));
	      AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	  }
      }
#endif
      if (ret == ceFormatNotSupported)
	  return ceFormatNotFound; /* makes prev work with unsupported files */
      if (ret != ceFormatNotFound)
	  return ret;
      cs.Seek(&cs, 0, SEEK_SET);
  }
  return PlayCurrentFileFast();
}
#endif/*USE_WAV*/

/*
  put packed string
 */
void putstrp(register __i0 u_int16 *packedStr) {
    while (1) {
	register int i = *packedStr++;
	if (i >> 8) {
	    putch(i >> 8);
	} else {
	    break;
	}
	if (i & 255) {
	    putch(i);
	} else {
	    break;
	}
    }
}

void IterateFilesCallback(register __b0 u_int16 *name) {
    register int i;
    for (i=0;i<32/2;i++) {
	putch(*name >> 8);
	putch(*name++);
    }
#ifdef SHOW_LONG_FILENAMES
    if (minifatInfo.longFileName[0]) {
	register __y s_int16 *p = minifatInfo.longFileName;
	while (*p >> 8) {
	    putch(*p >> 8);
	    if (*p & 255) {
		putch(*p & 255);
	    } else {
		break;
	    }
	    p++;
	}
    }
    putch('\n');
#endif
}

void main(void) {
#if defined(GPIO_CONTROL) && defined(GPIO_NUMBERED_FILES)
    u_int16 gpNum;
    u_int16 gpName[4] = "\pAUDIO00 ";
#endif

#ifdef EXTCLOCK
  extClock4KHz = EXTCLOCK/4;//12000/4;
#endif
  InitAudio();

  PERIP(INT_ENABLEL) |= INTF_RX | INTF_TIM0;
  PERIP(INT_ENABLEH) = INTF_DAC;
  
  PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA;
  PERIP(USB_CONFIG) = 0x8000U;

#if 0
  PERIP(GPIO1_ODATA) |=  LED1|LED2;
  PERIP(GPIO1_DDR)   |=  LED1|LED2;
  PERIP(GPIO1_MODE)  &= ~(LED1|LED2);
#endif

  player.volumeOffset = 0;
  player.pauseOn = 0;
#ifdef POFFTIME
  player.offDelay = POFFTIME;
#endif

  keyOld = KEY_POWER;
  keyOldTime = -32767;

  SetHookFunction((u_int16)OpenFile, Fat12OpenFile);
#ifdef GPIO_CONTROL
    SetHookFunction((u_int16)IdleHook, GPIOCtrlIdleHook);
#endif
 
#ifdef USE_WAV
  /* allow both .WAV and .OGG */
  supportedFiles = defSupportedFiles;
  /* Install the new file player function and return to player. */
  SetHookFunction((u_int16)PlayCurrentFile, PlayWavOrOggFile);
#else
  SetHookFunction((u_int16)PlayCurrentFile, PlayCurrentFileFast);
#endif

  // Use our SPI flash mapper as logical disk
  // In player mode we would only need to replace ReadDiskSector()
  map = FsMapSpiFlashCreate(NULL, 0); 
//  player.volume = 0;
//  PlayerVolume();

#if 1
  /* set cs.sampleRate to non-zero because we may use cs.Output() */
  //clockX = player.maxClock = 2; /* 1.0x */
  SetRate(cs.sampleRate = 8000U);

  /* These are for low-power pause mode, because USB mode is not entered */
  voltages[voltCoreSuspend] = 11; /*1.7V*/
  voltages[voltIoSuspend]   = 0;  /*1.7V*/
  voltages[voltAnaSuspend]  = 0;  /*2.4V*/

  /* If maxClock is restricted, can use lower voltages. */
  //voltages[voltCorePlayer] = 11; /*1.7V*/
  //voltages[voltIoPlayer]   = 0;  /*1.7V*/
  //voltages[voltAnaPlayer]  = 0;  /*2.4V*/
  //PowerSetVoltages(&voltages[voltCorePlayer]);
  //puts("power");

//  player.volume = 0;//+24;
//  PlayerVolume();
#endif
#ifdef PATCH_LOW_POWER_MODE
  SetHookFunction((u_int16)USBSuspend, MyUSBSuspend);
#endif
  
#ifdef GPIO_CONTROL
  GPIOInit();
#endif /*GPIO_CONTROL*/

  PlayerVolume();
  // Use custom main loop to take over total control of chip
  while (1) {
    //TODO: check if MMC/SD is inserted, otherwise
    CheckSd();

    // Try to use a FAT filesystem on logical disk
    if (InitFileSystem() == 0) {
	minifatInfo.supportedSuffixes = supportedFiles; //.ogg
      
	// Look for playable files
	player.totalFiles = OpenFile(0xffffU);
	if (player.totalFiles == 0) {
	    goto noFSnorFiles;
	}
	// Playable file(s) found. Play.
#if !defined(GPIO_CONTROL)
	player.nextFile = 0;
#endif
	while (1) {
	    CheckSd();

	    if (player.nextFile == -1) {
		player.currentFile = -1;
		while ( player.currentFile == -1 ) {

		    memset(tmpBuf, 0, sizeof(tmpBuf)); /* silence */
		    AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2); /*no pause*/
		    //cs.Output(&cs, tmpBuf, sizeof(tmpBuf)/2);

		    CheckSd();
		}
	    }

    playloop:
	    /* No random play */
	    player.currentFile = player.nextFile;
	    if (player.currentFile < 0) 
		player.currentFile += player.totalFiles;
	    if (player.currentFile >= player.totalFiles)
		player.currentFile -= player.totalFiles;

#if defined(GPIO_CONTROL)
	    player.nextFile = -1;
#else
	    player.nextFile = player.currentFile + 1;
#endif

//putch(0x55);
//putch(player.currentFile);
#if defined(GPIO_CONTROL) && defined(GPIO_NUMBERED_FILES)
	    /* numbered files AUDIO01.OGG .. AUDIOXX.OGG */
	    gpName[2] = (('O'<<8) | '0') + (player.currentFile / 10);
	    gpName[3] = ((player.currentFile % 10) << 8) + (('0'<<8) | ' ');
	    gpNum = OpenFileNamed(gpName, FAT_MKID('O','G','G'));
	    if (gpNum != 0xffffU)
#else
	    if (OpenFile(player.currentFile) < 0)
#endif
	    {
		player.ffCount = 0;
		cs.cancel = 0;
		cs.goTo = -1;
		cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
		cs.fastForward = 1;

		PlayCurrentFile(); // Decode and Play.
#ifdef POWER_OFF_AFTER_FILE
		PowerOff();
#endif

		if (USBIsAttached()) {
		    /* SPIplayer: Start over by watchdog reset. */
		    USEX(WDOG_CONFIG) = 10;
		    USEX(WDOG_RESET) = 0x4ea9; /*Activate watchdog*/
		    while (1)
			;
		}

	    } else {
		player.nextFile = 0;
	    }
	}
    } else {
    noFSnorFiles:
	LoadCheck(&cs, 32);
	memset(tmpBuf, 0, sizeof(tmpBuf)); /* silence */
	AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2); /* play silence */
    }
  }
}
